"""
davies.compass.plt: Module for parsing and working with Compass .PLT plot files
"""

import logging
import datetime

from davies.compass import ParseException, name_from_filename

log = logging.getLogger(__name__)

__all__ = 'Plot', 'Segment', 'MoveCommand', 'DrawCommand', 'CompassPltParser'


class Command(object):
    """Base class for Compass .PLT plot commands."""
    cmd = None

    def __init__(self, x, y, z, name, l, r, u, d, ele):
        self.x, self.y, self.z = x, y, z
        self.name = name
        self.l, self.r, self.u, self.d = l, r, u, d
        self.ele = ele


class MoveCommand(Command):
    """Compass .PLT plot command for moving the "plotting pen" to a specified X,Y,Z coordinate."""
    cmd = 'M'


class DrawCommand(Command):
    """Compass .PLT plot command for drawing a line segment between two points."""
    cmd = 'D'


class Segment(object):
    """Representation of a Compass .PLT segment. A Segment is a container for :class:`Command` objects."""

    def __init__(self, name=None, date=None, comment=None):
        self.name = name
        self.date = date
        self.comment = comment
        self.xmin = self.xmax = self.ymin = self.ymax = self.zmin = self.zmax = None
        self.commands = []

    def set_bounds(self, xmin, xmax, ymin, ymax, zmin, zmax):
        """Set X,Y,Z bounds for the segment."""
        self.xmin, self.xmax = xmin, xmax
        self.ymin, self.ymax = ymin, ymax
        self.zmin, self.zmax = zmin, zmax

    def add_command(self, command):
        """Add a :class:`Command` to :attr:`commands`."""
        self.commands.append(command)

    def __len__(self):
        return len(self.commands)

    def __iter__(self):
        for command in self.commands:
            yield command


class Plot(object):
    """Representation of a Compass .PLT plot file. A Plot is a container for :class:`Segment` objects."""

    def __init__(self, name=None):
        self.name = name
        self.utm_zone = None
        self.datum = None
        self.xmin = self.xmax = self.ymin = self.ymax = self.zmin = self.zmax = None
        self.segments = []
        self.fixed_points = {}  # name -> (x,y,z)

    def set_bounds(self, xmin, xmax, ymin, ymax, zmin, zmax):
        """Set X,Y,Z bounds for the plot."""
        self.xmin, self.xmax = xmin, xmax
        self.ymin, self.ymax = ymin, ymax
        self.zmin, self.zmax = zmin, zmax

    def add_segment(self, segment):
        """Add a :class:`Segment` to :attr:`segments`."""
        self.segments.append(segment)

    def add_fixed_point(self, name, coordinate):
        """Add an (X, Y, Z) tuple to :attr:`fixed_points`."""
        self.fixed_points[name] = coordinate

    def __len__(self):
        return len(self.segments)

    def __iter__(self):
        for segment in self.segments:
            yield segment

    def __contains__(self, item):
        for segment in self.segments:
            if item == segment.name:
                return True
        return False

    def __getitem__(self, item):
        for segment in self.segments:
            if item == segment.name:
                return segment
        raise KeyError(item)


class CompassPltParser(object):
    """Parser for Compass .PLT plot files."""
    # See:  http://www.fountainware.com/compass/Documents/FileFormats/PlotFileFormat.htm

    def __init__(self, pltfilename):
        """:param pltfilename: string filename"""
        self.pltfilename = pltfilename

    def parse(self):
        """Parse our .PLT file and return :class:`Plot` object or raise :exc:`ParseException`."""
        plt = Plot(name_from_filename(self.pltfilename))

        with open(self.pltfilename, 'rb') as pltfile:
            first_line = pltfile.readline()
            c, val = first_line[:1], first_line[1:]

            segment = None

            for line in pltfile:
                if not line:
                    continue

                c, val = line[:1], line[1:]

                if c == 'Z':
                    plt.set_bounds(*(float(v) for v in val.split()))

                if c == 'S':
                    # this is probably more appropriately called a "segment"
                    if not plt.name:
                        plt.name = val.strip()

                elif c == 'G':
                    plt.utm_zone = int(val)

                elif c == 'O':
                    plt.datum = val

                elif c == 'N':
                    name, _, m, d, y, comment = val.split(None, 5)
                    comment = comment[1:].strip()
                    date = datetime.date(int(y), int(m), int(d))
                    segment = Segment(name, date, comment)

                elif c == 'M':
                    x, y, z, name, _, l, u, d, r, _, ele = val.split()
                    cmd = MoveCommand(float(x), float(y), float(z), name[1:], float(l), float(r), float(u), float(d), float(ele))
                    segment.add_command(cmd)

                elif c == 'D':
                    x, y, z, name, _, l, u, d, r, _, ele = val.split()
                    cmd = DrawCommand(float(x), float(y), float(z), name[1:], float(l), float(r), float(u), float(d), float(ele))
                    segment.add_command(cmd)

                elif c == 'X':
                    segment.set_bounds(*(float(v) for v in val.split()))

                    # An X-bounds command signifies end of segment
                    plt.add_segment(segment)
                    segment = None

                elif c == 'P':
                    name, x, y, z = val.split()
                    plt.add_fixed_point(name, (float(x), float(y), float(z)))

                elif c == '\x1A':
                    continue  # "soft EOF" ascii SUB ^Z

                else:
                    raise ParseException("Unknown PLT control code '%s': %s" % (c, val))

        return plt



"""
davies.math: basic mathematics routines for reduction of survey data

This is "slow math", operating on scalar values without vector math (no `numpy` dependency).
"""

import math


__all__ = 'hd', 'vd', 'cartesian_offset', 'angle_delta'


def hd(inc, sd):
    """
    Calculate horizontal distance.

    :param inc: (float) inclination angle in degrees
    :param sd:  (float) slope distance in any units
    """
    return sd * math.cos(math.radians(inc))


def vd(inc, sd):
    """
    Calculate vertical distance.

    :param inc: (float) inclination angle in degrees
    :param sd:  (float) slope distance in any units
    """
    return abs(sd * math.sin(math.radians(inc)))


def cartesian_offset(azm, inc, sd, origin=(0, 0)):
    """
    Calculate the (X, Y) cartesian coordinate offset.

    :param azm:    (float) azimuth angle in degrees
    :param inc:    (float) inclination angle in degrees
    :param sd:     (float) slope distance in any units
    :param origin: (tuple(float, float)) optional origin coordinate
    """
    hd = sd * math.cos(math.radians(inc))
    x = hd * math.sin(math.radians(azm))
    y = hd * math.cos(math.radians(azm))
    return (x, y) if not origin else (x+origin[0], y+origin[1])


def angle_delta(a1, a2):
    """
    Calculate the absolute difference between two angles in degrees

    :param a1: (float) angle in degrees
    :param a2: (float) angle in degrees
    """
    return 180 - abs(abs(a1 - a2) - 180)



"""
davies.compass: Module for parsing and working with Compass source files
"""

import os.path
import logging
import datetime
import codecs
from collections import OrderedDict

log = logging.getLogger(__name__)

__all__ = 'Project', 'DatFile', 'Survey', 'Shot', 'UTMLocation', 'Exclude', \
          'CompassProjectParser', 'CompassDatParser', 'ParseException'


# Compass OO Model


class Exclude:
    """Shot flags"""
    LENGTH  = 'L'
    TOTAL   = 'X'
    CLOSURE = 'C'
    PLOT    = 'P'


class Shot(OrderedDict):
    """Representation of a single shot in a Compass Survey."""
    # FIXME: support compass, back-compass, and tape corrections

    def __init__(self, *args, **kwargs):
        """
        :kwarg FROM:    (str) from station
        :kwarg TO:      (str) to station
        :kwarg BEARING: (float) forward compass
        :kwarg AZM2:    (float) back compass
        :kwarg INC:     (float) forward inclination
        :kwarg INC2:    (float) back inclination
        :kwarg LENGTH:  (float) distance
        :kwarg FLAGS:   (collection of :class:`Exclude`) shot exclusion flags
        :kwarg COMMENTS: (str)
        :kwarg declination: (float) magnetic declination

        :ivar declination: (float) set or get magnetic declination adjustment
        """
        self.declination = kwargs.pop('declination', 0.0)
        OrderedDict.__init__(self, *args, **kwargs)

    @property
    def azm(self):
        """Corrected azimuth, taking into account backsight, declination, and compass corrections."""
        azm1 = self.get('BEARING', None)
        azm2 = self.get('AZM2', None)
        if azm1 is None and azm2 is None:
            return None
        if azm2 is None:
            return azm1 + self.declination
        if azm1 is None:
            return (azm2 + 180) % 360 + self.declination
        return (azm1 + (azm2 + 180) % 360) / 2.0 + self.declination

    @property
    def inc(self):
        """Corrected inclination, taking into account backsight and clino corrections."""
        inc1 = self.get('INC', None)
        inc2 = self.get('INC2', None)
        if inc1 is None and inc2 is None:
            return None
        if inc2 is None:
            return inc1
        if inc1 is None:
            return -1 * inc2
        return (inc1 - inc2) / 2.0

    @property
    def flags(self):
        """Shot exclusion flags as a `set`"""
        return set(self.get('FLAGS', ''))  # older data may not have FLAGS field

    @property
    def length(self):
        """Corrected distance, taking into account tape correction."""
        return self.get('LENGTH', None)

    @property
    def is_included(self):
        return Exclude.LENGTH not in self.flags and Exclude.TOTAL not in self.flags

    @property
    def is_excluded(self):
        return Exclude.LENGTH in self.flags or Exclude.TOTAL in self.flags

    def __str__(self):
        return ', '.join('%s=%s' % (k,v) for (k,v) in self.items())

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self)


class Survey(object):
    """Representation of a Compass Survey object. A Survey is a container for :class:`Shot` objects."""

    def __init__(self, name=None, date=None, comment=None, team=None, declination=None, lrud_format=None, corrections=None, corrections2=None, cave_name=None, shot_header=None, shots=None):
        self.name = name
        self.date = date
        self.comment = comment
        self.team = team
        self.declination = declination
        self.lrud_format = lrud_format  # TODO: LRUD and units not supported
        self.corrections, self.corrections2 = corrections, corrections2  # TODO: instrument corrections not supported
        self.cave_name = cave_name
        self.shot_header = shot_header  # FIXME: this ordering is not optional!
        self.shots = shots if shots else []

    def add_shot(self, shot):
        """Add a shot dictionary to :attr:`shots`, applying this survey's magnetic declination"""
        shot.declination = self.declination
        self.shots.append(shot)

    @property
    def length(self):
        """Total surveyed length, regardless of exclusion flags."""
        return sum([shot.length for shot in self.shots])

    @property
    def included_length(self):
        """Surveyed length, not including "excluded" shots"""
        return sum([shot.length for shot in self.shots if shot.is_included])

    @property
    def excluded_length(self):
        """Surveyed length which does not count toward the included total"""
        return sum([shot.length for shot in self.shots if Exclude.LENGTH in shot.flags or Exclude.TOTAL in shot.flags])

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)

    def __len__(self):
        return len(self.shots)

    def __iter__(self):
        for shot in self.shots:
            yield shot

    def __contains__(self, item):
        for shot in self.shots:
            if item in (shot.get('FROM', None), shot.get('TO', None)):
                return True
        return False

    def _serialize(self):
        lines = [
            self.cave_name,
            'SURVEY NAME: %s' % self.name,
            'SURVEY DATE: %s  COMMENT:%s' % (self.date.strftime('%-m %d %Y'), self.comment),
            'SURVEY TEAM:',
            ','.join(self.team) if self.team else '',
            'DECLINATION: %7.2f  FORMAT: %s  CORRECTIONS:  %s  CORRECTIONS2:  %s' %
                (self.declination, self.lrud_format or '',
                 '%.2f %.2f %.2f' % tuple(self.corrections),
                 '%.2f %.2f' % tuple(self.corrections2)),
            '',
            '\t'.join(self.shot_header),
            '',
        ]
        for shot in self.shots:
            vals = []
            for k,v in shot.items():  # FIXME: this depends on OrderedDict ordering, ensure we match `shot_header`
                if k in ('BEARING', 'INC', 'AZM2', 'INC2', 'LENGTH'):
                    vals.append('%7.2f' % (v if v is not None else -999.0))
                elif k in ('LEFT', 'RIGHT', 'UP', 'DOWN'):
                    vals.append('%7.2f' % (v if v is not None and v != float('inf') else -9.90))
                elif k in ('FROM', 'TO'):
                    v = v or ''
                    vals.append(v.rjust(6))
                elif k in ('FLAGS', 'COMMENTS'):
                    pass  # handle them together below
                else:
                    vals.append(str(v) if v is not None else '')

            if shot.get('FLAGS', ''):
                flags = '#|%s#' % ''.join(list(shot.get('FLAGS', ())))
                vals.append('%s  %s' % (flags, shot.get('COMMENTS', '') or ''))
            else:
                vals.append((shot.get('COMMENTS', '') or '')[:80])

            lines.append('\t'.join(vals))
        return lines


class DatFile(object):
    """
    Representation of a Compass .DAT File. A DatFile is a container for :class:`Survey` objects.

    :ivar name:    (string) the DatFile's "name", not necessarily related to its filename
    :ivar surveys: (list of :class:`Survey`)
    """

    def __init__(self, name=None):
        self.name = name
        self.surveys = []

    def add_survey(self, survey):
        """Add a :class:`Survey` to :attr:`surveys`."""
        self.surveys.append(survey)

    @property
    def length(self):
        """Total surveyed length."""
        return sum([survey.length for survey in self.surveys])

    @property
    def included_length(self):
        """Surveyed length, not including "excluded" shots"""
        return sum([survey.included_length for survey in self.surveys])

    @property
    def excluded_length(self):
        """Surveyed length which does not count toward the included total"""
        return sum([survey.excluded_length for survey in self.surveys])

    def __len__(self):
        return len(self.surveys)

    def __iter__(self):
        for survey in self.surveys:
            yield survey

    def __contains__(self, item):
        for survey in self.surveys:
            if item == survey.name or item == survey:
                return True
        return False

    def __getitem__(self, item):
        for survey in self.surveys:
            if item == survey.name or item == survey:
                return survey
        raise KeyError(item)

    @staticmethod
    def read(fname):
        """Read a .DAT file and produce a `Survey`"""
        return CompassDatParser(fname).parse()

    def write(self, outf):
        """Write a `Survey` to the specified .DAT file"""
        with codecs.open(outf, 'wb', 'windows-1252') as outf:
            for survey in self.surveys:
                outf.write('\r\n'.join(survey._serialize()))
                outf.write('\r\n'+'\f'+'\r\n')  # ASCII "form feed" ^L
            outf.write('\x1A')  # ASCII "sub" ^Z marks EOF


class UTMLocation(object):
    """Represents a UTM-based coordinate for fixed stations."""

    def __init__(self, easting, northing, elevation, zone=None, convergence=None, datum=None):
        self.easting = easting
        self.northing = northing
        self.elevation = elevation
        self.zone = int(zone) if zone is not None else None
        self.convergence = convergence
        self.datum = datum

    @property
    def __geo_interface__(self):
        return {'type': 'Point', 'coordinates': (self.easting, self.northing, self.elevation)}

    def __str__(self):
        return "<%s UTM Zone %s %0.1fE %0.1fN %0.1f>" % (self.datum, self.zone, self.easting, self.northing, self.elevation)


class Project(object):
    """
    Representation of a Compass .MAK Project file. A Project is a container for :class:`DatFile` objects.

    :ivar name:         (string)
    :ivar linked_files: (list of :class:`DatFile`)
    """

    def __init__(self, name=None, base_location=None, file_params=None):
        self.name = name
        self.base_location = base_location
        self.file_params = file_params
        self.linked_files = []

    def add_linked_file(self, datfile):
        """Add a :class:`DatFile` to :attr:`linked_files`."""
        self.linked_files.append(datfile)

    def __len__(self):
        return len(self.linked_files)

    def __iter__(self):
        for linked_file in self.linked_files:
            yield linked_file

    def __getitem__(self, item):
        for datfile in self.linked_files:
            if item == datfile.name or item == datfile:
                return datfile
        raise KeyError(item)

    @staticmethod
    def read(fname):
        """Read a .MAK file and produce a `Project`"""
        return CompassProjectParser(fname).parse()

    def write(self, outf):
        raise NotImplementedError()



# File Parsing Utilities


_FLOAT_KEYS = ['LENGTH', 'BEARING', 'AZM2', 'INC', 'INC2', 'LEFT', 'RIGHT', 'UP', 'DOWN']
_INF_KEYS = ['LEFT', 'RIGHT', 'UP', 'DOWN']


def name_from_filename(fname):
    return os.path.splitext(os.path.basename(fname))[0].replace('_', ' ')


class ParseException(Exception):
    """Exception raised when parsing fails."""
    pass


class CompassSurveyParser(object):
    """Parser for a Compass survey string."""

    def __init__(self, survey_str):
        """:param survey_str: string multiline representation of survey as found in .DAT file"""
        self.survey_str = survey_str

    @staticmethod
    def _coerce(key, val):
        if val == '-999.00':  # no data
            return None

        if key in _INF_KEYS and val in ('-9.90', '-9999.00'):  # passage
            return float('inf')

        if key in _FLOAT_KEYS:
            try:
                return float(val)
            except TypeError as e:
                log.warn("Unable to coerce to float %s=%s (%s)", key, val, type(val))

        return val

    @staticmethod
    def _parse_date(datestr):
        datestr = datestr.strip()
        for fmt in ['%m %d %Y', '%m %d %y']:
            try:
                return datetime.datetime.strptime(datestr, fmt).date()
            except ValueError:
                pass
        raise ParseException("Unable to parse SURVEY DATE: %s" % datestr)

    @staticmethod
    def _parse_declination_line(line):
        declination, fmt, corrections, corrections2 = 0.0, '', (0.0, 0.0, 0.0), (0.0, 0.0)
        toks = line.strip().split()
        for i, tok in enumerate(toks):
            if tok == 'DECLINATION:':
                declination = float(toks[i+1])
            elif tok == 'FORMAT:':
                fmt = toks[i+1]
            elif tok == 'CORRECTIONS:':
                corrections = map(float, toks[i+1:i+4])
            elif tok == 'CORRECTIONS2:':
                corrections2 = map(float, toks[i+1:i+3])
        return declination, fmt, corrections, corrections2

    def parse(self):
        """Parse our string and return a Survey object, None, or raise :exc:`ParseException`"""
        if not self.survey_str:
            return None
        lines = self.survey_str.splitlines()
        if len(lines) < 10:
            raise ParseException("Expected at least 10 lines in a Compass Survey, only found %d!\nlines=%s" % (len(lines), lines))

        # undelimited Cave Name may be empty string and "skipped"
        first_line = lines.pop(0).strip()
        if first_line.startswith('SURVEY NAME:'):
            cave_name = ''
            name = first_line.strip('SURVEY NAME:').strip()
        else:
            cave_name = first_line
            name = lines.pop(0).split('SURVEY NAME:', 1)[1].strip()

        # Date and Comment on one line, Comment may be missing
        date_comment_toks = lines.pop(0).split('SURVEY DATE:', 1)[1].split('COMMENT:')
        date = CompassSurveyParser._parse_date(date_comment_toks[0])
        comment = date_comment_toks[1].strip() if len(date_comment_toks) > 1 else ''

        lines.pop(0)  # SURVEY TEAM:\n (actual team members are on the next line)
        team = [member.strip() for member in lines.pop(0).split(',')]  # We're already decoding from windows-1252 codec so we have unicode for names like 'Tanya Pietra\xdf'

        # TODO: implement format (units!), instrument correction(s)
        dec_fmt_corr = lines.pop(0)
        declination, fmt, corrections, corrections2 = CompassSurveyParser._parse_declination_line(dec_fmt_corr)

        lines.pop(0)
        shot_header = lines.pop(0).split()
        val_count = len(shot_header) - 2 if 'FLAGS' in shot_header else len(shot_header)  # 1998 vintage data has no FLAGS, COMMENTS at end
        lines.pop(0)

        survey = Survey(name=name, date=date, comment=comment, team=team, cave_name=cave_name,
                        shot_header=shot_header, declination=declination,
                        lrud_format=fmt, corrections=corrections, corrections2=corrections2)

        shots = []
        shot_lines = lines
        for shot_line in shot_lines:
            shot_vals = shot_line.split(None, val_count)

            if len(shot_vals) > val_count:  # last two spare columns are FLAGS and COMMENTS, either value may be missing
                flags_comment = shot_vals.pop()
                if not flags_comment.startswith('#|'):
                    flags, comment = '', flags_comment
                else:
                    try:
                        flags, comment = flags_comment.split('#|', 1)[1].split('#', 1)
                    except ValueError:
                        raise ParseException('Invalid flags in %s survey: %s' % (name, flags_comment))  # A 2013 bug in Compass inserted corrupt binary garbage into FLAGS column, causes parse to barf
                shot_vals += [flags, comment.strip()]

            shot_vals = [(header, self._coerce(header, val)) for (header, val) in zip(shot_header, shot_vals)]
            shot = Shot(shot_vals)
            survey.add_shot(shot)

        #log.debug("Survey: name=%s shots=%d length=%0.1f date=%s team=%s\n%s", name, len(shots), survey.length, date, team, '\n'.join([str(shot) for shot in survey.shots]))

        return survey


class CompassDatParser(object):
    """Parser for Compass .DAT data files"""

    def __init__(self, datfilename):
        """:param datfilename: (string) filename"""
        self.datfilename = datfilename

    def parse(self):
        """Parse our data file and return a :class:`DatFile` or raise :exc:`ParseException`."""
        log.debug("Parsing Compass .DAT file %s ...", self.datfilename)
        datobj = DatFile(name_from_filename(self.datfilename))

        with codecs.open(self.datfilename, 'rb', 'windows-1252') as datfile:
            full_contents = datfile.read()
            survey_strs = [survey_str.strip() for survey_str in full_contents.split('\x0C')]

            if survey_strs[-1] == '\x1A':
                survey_strs.pop()  # Compass may place a "soft EOF" with ASCII SUB char

            log.debug("Parsed %d raw surveys from Compass .DAT file %s.", len(survey_strs), self.datfilename)
            for survey_str in survey_strs:
                if not survey_str:
                    continue
                survey = CompassSurveyParser(survey_str).parse()
                datobj.add_survey(survey)

        return datobj


class CompassProjectParser(object):
    """Parser for Compass .MAK project files."""

    def __init__(self, projectfile):
        """:param projectfile: (string) filename"""
        self.makfilename = projectfile

    def parse(self):
        """Parse our project file and return :class:`Project` object or raise :exc:`ParseException`."""
        log.debug("Parsing Compass .MAK file %s ...", self.makfilename)

        base_location = None
        linked_files = []
        file_params = set()

        def parse_linked_file(value):
            log.debug("Parsing linked file:  %s", value)
            value = value.rstrip(';')
            toks = value.split(',', 1)
            if len(toks) == 1:
                return toks[0]
            else:
                return toks[0]  # TODO: implement link stations and fixed stations

        with codecs.open(self.makfilename, 'rb', 'windows-1252') as makfile:
            prev = None

            for line in makfile:
                line = line.strip()

                if not line:
                    continue

                header, value = line[0], line[1:]

                if prev:
                    if line.endswith(';'):
                        linked_file = parse_linked_file(prev + line.rstrip(';'))
                        linked_files.append(linked_file)
                        prev = None
                    else:
                        prev += value
                    continue

                if header == '/':
                    pass  # comment

                elif header == '@':
                    value = value.rstrip(';')
                    base_location = UTMLocation(*(float(v) for v in value.split(',')))

                elif header == '&':
                    value = value.rstrip(';')
                    base_location.datum = value

                elif header == '%':
                    value = value.rstrip(';')
                    base_location.convergence = float(value)

                elif header == '!':
                    value = value.rstrip(';')
                    file_params = set(value.upper())

                elif header == '#':
                    if value.endswith(';'):
                        linked_files.append(parse_linked_file(value))
                        prev = None
                    else:
                        prev = value

            log.debug("Project:  base_loc=%s  params=%s  linked_files=%s", base_location, file_params, linked_files)

            project = Project(name_from_filename(self.makfilename), base_location, file_params)

            for linked_file in linked_files:
                # TODO: we need to support case-insensitive path resolution on case-sensitive filesystems
                linked_file_path = os.path.join(os.path.dirname(self.makfilename), os.path.normpath(linked_file.replace('\\', '/')))
                datfile = CompassDatParser(linked_file_path).parse()
                project.add_linked_file(datfile)

            return project



"""
davies.pockettopo: Module for parsing and working with exported PocketTopo survey data
"""

from __future__ import division
from __future__ import print_function

import re
import codecs
import logging
from datetime import datetime
from collections import OrderedDict, defaultdict

log = logging.getLogger(__name__)

__all__ = 'TxtFile', 'Survey', 'MergingSurvey', 'Shot', 'PocketTopoTxtParser'


# TODO: properly handle zero-length shots with both from/to (station equivalence)
# TODO: older versions didn't specify units?


class Shot(OrderedDict):
    """
    Representation of a single shot in a PocketTopo Survey.

    :kwarg FROM:    (str) from station
    :kwarg TO:      (str) optional to station
    :kwarg LENGTH:  (float) distance
    :kwarg AZM:     (float) compass
    :kwarg INC:     (float) inclination
    :kwarg COMMENT: (str)
    :kwarg declination: (float) optional

    :ivar declination: (float) set or get the applied magnetic declination for the shot
    """

    def __init__(self, *args, **kwargs):
        self.declination = kwargs.pop('declination', 0.0)
        OrderedDict.__init__(self, *args, **kwargs)

        self.dupe_count = 1  # denotes averaged backsights (2) and triple-shots (3)

    @property
    def azm(self):
        """Corrected azimuth, taking into account declination."""
        return self.get('AZM', -0.0) + self.declination

    @property
    def inc(self):
        """Corrected inclination."""
        return self.get('INC', -0.0)

    @property
    def length(self):
        """Corrected distance."""
        return self.get('LENGTH', -0.0)

    @property
    def is_splay(self):
        """Is this shot a "splay shot"?"""
        return self.get('TO', None) in (None, '')

    def __str__(self):
        return ', '.join('%s=%s' % (k,v) for (k,v) in self.items())

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self)


class Survey(object):
    """
    Representation of a PocketTopo Survey object. A Survey is a container for :class:`Shot` objects.
    """

    def __init__(self, name=None, date=None, comment=None, declination=0.0, cave_name=None, length_units='m', angle_units=360, shots=None):
        self.name = name
        self.date = date
        self.comment = comment
        self.declination = declination
        self.cave_name = cave_name
        self.length_units = length_units
        self.angle_units = angle_units

        self.shots = []
        self.splays = defaultdict(list)
        if shots:
            [self.add_shot(shot) for shot in shots]

    def add_shot(self, shot):
        """Add a Shot to :attr:`shots`, applying our survey's :attr:`declination` to it."""
        shot.declination = self.declination
        if shot.is_splay:
            self.splays[shot['FROM']].append(shot)
        self.shots.append(shot)

    @property
    def length(self):
        """Total surveyed cave length, not including splays."""
        return sum([shot.length for shot in self.shots if not shot.is_splay])

    @property
    def total_length(self):
        """Total surveyed length including splays."""
        return sum([shot.length for shot in self.shots])

    def __len__(self):
        return len(self.shots)

    def __iter__(self):
        for shot in self.shots:
            yield shot

    def __contains__(self, item):
        for shot in self.shots:
            if item in (shot.get('FROM', None), shot.get('TO', None)):
                return True
        return False

    def __str__(self):
        return self.name

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.name)

    # def _serialize(self):
    #     return []


class MergingSurvey(Survey):
    """
    Representation of a PocketTopo Survey object. A Survey is a container for :class:`Shot` objects.
    This Survey implementation merges "duplicate" shots into a single averaged shot.

    PocketTopo (and DistoX) convention is to use triple forward shots for mainline survey. When
    adding a new shot to this class with `add_shot()`, if we detect that the previous shot was
    between the same two stations, we average values and merge the two together instead of appending
    the duplicate shot. We use a "running" mean algorithm, so that this feature works for any number
    of subsequent duplicate shots (two, three, four...).
    """
    # For performance, we only look backwards at the immediately preceding shots!

    def _inverse_azm(self, azm):
        """Convert forward AZM to back AZM and vice versa"""
        return (azm + self.angle_units/2) % self.angle_units

    def _inverse_inc(self, inc):
        """Convert forward INC to back INC and vice versa"""
        return -1 * inc

    def add_shot(self, shot):
        """
        Add a shot dictionary to :attr:`shots`, applying our survey's :attr:`declination`, and
        optionally averaging and merging with duplicate previous shot.
        """
        if not self.shots or not shot.get('TO', None) or not self.shots[-1].get('TO', None):
            return super(MergingSurvey, self).add_shot(shot)

        from_, to = shot['FROM'], shot['TO']
        prev_shot = self.shots[-1]
        prev_from, prev_to = prev_shot['FROM'], prev_shot['TO']

        if from_ == prev_from and to == prev_to:
            # dupe shot! calculate iterative "running" mean and merge into the previous shot
            total_count = prev_shot.dupe_count + 1

            log.debug('Merging %d shots "%s" <- "%s"', total_count, prev_shot, shot)
            if abs(shot['AZM'] - prev_shot['AZM']) > 2.0:
                log.warning('Merged forward AZM disagreement of %0.1f for "%s" <- "%s"', abs(shot['AZM'] - prev_shot['AZM']), prev_shot, shot)
            if abs(shot['INC'] - prev_shot['INC']) > 2.0:
                log.warning('Merged forward INC disagreement of %0.1f for "%s" <- "%s"', abs(shot['INC'] - prev_shot['INC']), prev_shot, shot)
            if abs(shot['LENGTH'] - prev_shot['LENGTH']) > 1.0:
                log.warning('Merged forward LENGTH disagreement of %0.1f for "%s" <- "%s"', abs(shot['LENGTH'] - prev_shot['LENGTH']), prev_shot, shot)

            avg_length = (prev_shot['LENGTH'] * prev_shot.dupe_count + shot['LENGTH']) / total_count
            avg_azm = (prev_shot['AZM'] * prev_shot.dupe_count + shot['AZM']) / total_count
            avg_inc = (prev_shot['INC'] * prev_shot.dupe_count + shot['INC']) / total_count
            merged_comments = ('%s %s' % (prev_shot.get('COMMENT', '') or '', shot.get('COMMENT', '') or '')).strip() or None

            prev_shot['LENGTH'], prev_shot['AZM'], prev_shot['INC'], prev_shot['COMMENT'] = avg_length, avg_azm, avg_inc, merged_comments
            prev_shot.dupe_count += 1

        elif from_ == prev_to and to == prev_from:
            # backsight! we do the same iterative "running" mean rather than assuming a single forward and single back
            total_count = prev_shot.dupe_count + 1
            inv_azm, inv_inc = self._inverse_azm(shot['AZM']), self._inverse_inc(shot['INC'])

            log.debug('Merging %d backsights "%s" <- "%s"', total_count, prev_shot, shot)
            if abs(inv_azm - prev_shot['AZM']) > 2.0:
                log.warning('Backsight AZM disagreement of %0.1f for "%s" <- "%s"', abs(inv_azm - prev_shot['AZM']), prev_shot, shot)
            if abs(inv_inc - prev_shot['INC']) > 2.0:
                log.warning('Backsight INC disagreement of %0.1f for "%s" <- "%s"', abs(inv_inc - prev_shot['INC']), prev_shot, shot)
            if abs(shot['LENGTH'] - prev_shot['LENGTH']) > 1.0:
                log.warning('Backsight LENGTH disagreement of %0.1f for "%s" <- "%s"', abs(shot['LENGTH'] - prev_shot['LENGTH']), prev_shot, shot)

            avg_length = (prev_shot['LENGTH'] * prev_shot.dupe_count + shot['LENGTH']) / total_count
            avg_azm = (prev_shot['AZM'] * prev_shot.dupe_count + inv_azm) / total_count
            avg_inc = (prev_shot['INC'] * prev_shot.dupe_count + inv_inc) / total_count
            merged_comments = ('%s %s' % (prev_shot.get('COMMENT', '') or '', shot.get('COMMENT', '') or '')).strip() or None

            prev_shot['LENGTH'], prev_shot['AZM'], prev_shot['INC'], prev_shot['COMMENT'] = avg_length, avg_azm, avg_inc, merged_comments
            prev_shot.dupe_count += 1

        else:
            # a new, different shot; no merge
            return super(MergingSurvey, self).add_shot(shot)


class UTMLocation(object):
    """
    Represents a UTM-based coordinate for Reference Point.

    Note that PocketTopo doesn't support UTM Zones.

    :ivar easting:    (float)
    :ivar northing:   (float)
    :ivar elevation:  (float) meters
    :ivar comment:    (str)
    """

    def __init__(self, easting, northing, elevation=0.0, comment=None):
        self.easting = easting
        self.northing = northing
        self.elevation = elevation
        self.altitude = elevation  # alias
        self.comment = comment

    @property
    def __geo_interface__(self):
        return {'type': 'Point', 'coordinates': (self.easting, self.northing, self.elevation)}

    def __str__(self):
        return "<UTM %0.1fE %0.1fN %0.1fm>" % (self.easting, self.northing, self.elevation)


class TxtFile(object):
    """
    Representation of a PocketTopo .TXT File. A TxtFile is a container for :class:`Survey` objects.

    :ivar name:          (string) the TxtFile's "name"
    :ivar length_units:  (string) `m` (default) or `feet`
    :ivar angle_units:   (int) `360` for degrees (default) or `400` for grads
    :ivar surveys:       (list of :class:`Survey`)
    :ivar reference_points:  (dict of :class:`UTMLocation` by station)
    """

    def __init__(self, name=None, length_units='m', angle_units=360):
        self.name = name

        if length_units not in ('m', 'feet'):
            raise Exception('Length units must be either \'m\' for meters (default) or \'feet\' for feet')
        self.length_units = length_units

        if angle_units not in (360, '360', 400, '400'):
            raise Exception('Angle units must be either `360` for degrees (default) or `400` for grads')
        self.angle_units = int(angle_units)

        self.surveys = []
        self.reference_points = OrderedDict()

    def add_survey(self, survey):
        """Add a :class:`Survey` to :attr:`surveys`."""
        survey.length_units = self.length_units
        survey.angle_units = self.angle_units
        self.surveys.append(survey)

    def add_reference_point(self, station, utm_location):
        """Add a :class:`UTMLocation` to :attr:`reference_points`."""
        self.reference_points[station] = utm_location

    @property
    def length(self):
        """Total surveyed length."""
        return sum([survey.length for survey in self.surveys])

    def __len__(self):
        return len(self.surveys)

    def __iter__(self):
        for survey in self.surveys:
            yield survey

    def __contains__(self, item):
        for survey in self.surveys:
            if item == survey.name or item == survey:
                return True
        return False

    def __getitem__(self, item):
        for survey in self.surveys:
            if item == survey.name or item == survey:
                return survey
        raise KeyError(item)

    @staticmethod
    def read(fname, merge_duplicate_shots=False, encoding='windows-1252'):
        """Read a PocketTopo .TXT file and produce a `TxtFile` object which represents it"""
        return PocketTopoTxtParser(fname, merge_duplicate_shots, encoding).parse()

    # def write(self, outf):
    #     """Write a `Survey` to the specified .DAT file"""
    #     with codecs.open(outf, 'wb', 'windows-1252') as outf:
    #         for survey in self.surveys:
    #             outf.write('\r\n'.join(survey._serialize()))


class PocketTopoTxtParser(object):
    """Parses the PocketTopo .TXT file format"""

    def __init__(self, txtfilename, merge_duplicate_shots=False, encoding='windows-1252'):
        self.txtfilename = txtfilename
        self.merge_duplicate_shots = merge_duplicate_shots
        self.encoding = encoding

    def parse(self):
        """Produce a `TxtFile` object from the .TXT file"""
        log.debug('Parsing PocketTopo .TXT file %s ...', self.txtfilename)
        SurveyClass = MergingSurvey if self.merge_duplicate_shots else Survey
        txtobj = None

        with codecs.open(self.txtfilename, 'rb', self.encoding) as txtfile:
            lines = txtfile.read().splitlines()

            # first line is cave name and units
            first_line_re = re.compile(r'^([\w\s]*)\(([\w\s]*),([\w\s]*)')
            first_line = lines.pop(0)
            cave_name, length_units, angle_units = first_line_re.search(first_line).groups()
            cave_name, angle_units = cave_name.strip(), int(angle_units)
            txtobj = TxtFile(cave_name, length_units, angle_units)

            while not lines[0]:
                lines.pop(0)  # skip blanks

            # next block identifies surveys (trip) metadata
            while lines[0].startswith('['):
                toks = lines.pop(0).split(None, 3)
                id, date, declination = toks[:3]
                id = id.strip('[]:')
                date = datetime.strptime(date, '%Y/%m/%d').date()
                declination = float(declination)
                comment = toks[3].strip('"') if len(toks) == 4 else ''
                survey = SurveyClass(id, date, comment, declination, cave_name)
                txtobj.add_survey(survey)

            while not lines[0]:
                lines.pop(0)  # skip blanks

            # finally actual survey data
            while lines:
                line = lines.pop(0).strip()
                if not line:
                    continue

                if '"' in line:
                    line, comment = line.split('"', 1)
                    comment = comment.rstrip('"')
                else:
                    comment = None

                if '[' not in line:
                    # this is either a Reference Point or a zero-length fake shot
                    toks = line.split()
                    if len(toks) != 4:  # ??
                        log.debug('Skipping unrecognized shot:  %s %s', line, '"%s"' % comment if comment else '')
                        continue
                    station, vals = toks[0], list(map(float, toks[1:]))
                    if vals[0] == 0.0:  # fake shot
                        log.debug('Skipping zero-length shot:  %s %s', line, '"%s"' % comment if comment else '')
                    else:  # reference point
                        easting, northing, altitude = vals
                        reference_point = UTMLocation(easting, northing, altitude, comment)
                        log.debug('Reference point:  %s', reference_point)
                        txtobj.add_reference_point(station, reference_point)
                    continue

                line, survey_id = line.split('[')
                survey_id = survey_id.rstrip().rstrip(']')
                toks = line.split()
                from_to, (length, azm, inc) = toks[:-3], (float(tok) for tok in toks[-3:])

                if len(from_to) == 2:
                    from_, to = tuple(from_to)  # shot
                elif len(from_to) == 1:
                    from_, to = from_to[0], None  # splay
                elif not from_to and length == 0.0:
                    continue  # skip junk zero-length placeholder shots
                else:
                    raise Exception()

                shot = Shot([('FROM',from_), ('TO',to), ('LENGTH',length), ('AZM',azm), ('INC',inc), ('COMMENT',comment)])
                txtobj[survey_id].add_shot(shot)

        return txtobj


if __name__ == '__main__':
    import sys

    logging.basicConfig(level=logging.DEBUG)

    for fname in sys.argv[1:]:
        txtfile = PocketTopoTxtParser(fname, merge_duplicate_shots=True).parse()
        print('%s  (%s, %d)' % (txtfile.name, txtfile.length_units, txtfile.angle_units))
        for survey in txtfile:
            print('\t', '[%s] %s (%0.1f %s)' % (survey.name, survey.comment, survey.length, txtfile.length_units))
            for shot in survey:
                print('\t\t', shot)
